from enum import Enum
from typing import Any

from recap.converters.dbapi import DbapiConverter
from recap.types import (
    BytesType,
    FloatType,
    IntType,
    NullType,
    RecapType,
    StringType,
    UnionType,
)

# SQLite's maximum length is 2^31-1 bytes, or 2147483647 bytes.
SQLITE_MAX_LENGTH = 2147483647


class SQLiteAffinity(Enum):
    """
    SQLite uses column affinity to map non-STRICT table columns to values. See
    https://www.sqlite.org/datatype3.html#type_affinity for details.
    """

    INTEGER = "integer"
    REAL = "real"
    TEXT = "text"
    BLOB = "blob"
    NUMERIC = "numeric"


class SQLiteConverter(DbapiConverter):
    def _parse_type(self, column_props: dict[str, Any]) -> RecapType:
        column_name = column_props["COLUMN_NAME"]
        column_type = column_props["TYPE"]
        octet_length = column_props["CHARACTER_OCTET_LENGTH"]
        precision = column_props["NUMERIC_PRECISION"]

        match SQLiteConverter.get_affinity(column_type):
            case SQLiteAffinity.INTEGER:
                return IntType(bits=64)
            case SQLiteAffinity.REAL:
                if precision and precision <= 23:
                    return FloatType(bits=32)
                return FloatType(bits=64)
            case SQLiteAffinity.TEXT:
                return StringType(bytes_=octet_length or SQLITE_MAX_LENGTH)
            case SQLiteAffinity.BLOB:
                return BytesType(bytes_=octet_length or SQLITE_MAX_LENGTH)
            case SQLiteAffinity.NUMERIC:
                # NUMERIC affinity may contain values using all five storage classes
                return UnionType(
                    types=[
                        NullType(),
                        IntType(bits=64),
                        FloatType(bits=64),
                        StringType(bytes_=SQLITE_MAX_LENGTH),
                        BytesType(bytes_=SQLITE_MAX_LENGTH),
                    ]
                )
            case _:
                raise ValueError(
                    f"Unsupported `{column_type}` type for `{column_name}`"
                )

    @staticmethod
    def get_affinity(column_type: str | None) -> SQLiteAffinity:
        """
        Encode affinity rules as defined here:

        https://www.sqlite.org/datatype3.html#determination_of_column_affinity

        :param column_type: The column type to determine the affinity of.
        :return: The affinity of the column type.
        """

        column_type = (column_type or "").upper()

        if not column_type:
            return SQLiteAffinity.BLOB
        elif "INT" in column_type:
            return SQLiteAffinity.INTEGER
        elif "CHAR" in column_type or "TEXT" in column_type or "CLOB" in column_type:
            return SQLiteAffinity.TEXT
        elif "BLOB" in column_type:
            return SQLiteAffinity.BLOB
        elif "REAL" in column_type or "FLOA" in column_type or "DOUB" in column_type:
            return SQLiteAffinity.REAL
        else:
            return SQLiteAffinity.NUMERIC



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from recap.clients.dbapi import Connection, DbapiClient
from recap.converters.postgresql import PostgresqlConverter
from recap.types import StructType

PSYCOPG2_CONNECT_ARGS = {
    "host",
    "hostaddr",
    "port",
    "dbname",
    "user",
    "password",
    "passfile",
    "channel_binding",
    "connect_timeout",
    "client_encoding",
    "options",
    "application_name",
    "fallback_application_name",
    "keepalives",
    "keepalives_idle",
    "keepalives_interval",
    "keepalives_count",
    "tcp_user_timeout",
    "replication",
    "gssencmode",
    "sslmode",
    "requiressl",
    "sslcompression",
    "sslcert",
    "sslkey",
    "sslpassword",
    "sslrootcert",
    "sslcrl",
    "sslcrldir",
    "sslsni",
    "requirepeer",
    "ssl_min_protocol_version",
    "ssl_max_protocol_version",
    "krbsrvname",
    "gsslib",
    "service",
    "target_session_attrs",
}


class PostgresqlClient(DbapiClient):
    def __init__(
        self,
        connection: Connection,
        converter: PostgresqlConverter = PostgresqlConverter(),
    ) -> None:
        super().__init__(connection, converter)

    @staticmethod
    @contextmanager
    def create(
        paths: list[str] | None = None,
        **url_args,
    ) -> Generator[PostgresqlClient, None, None]:
        import psycopg2

        if paths:
            url_args["dbname"] = paths[0]

        # Only include kwargs that are valid for PsycoPG2 parse_dsn()
        url_args = {k: v for k, v in url_args.items() if k in PSYCOPG2_CONNECT_ARGS}

        with psycopg2.connect(**url_args) as client:
            yield PostgresqlClient(client)

    def ls_catalogs(self) -> list[str]:
        cursor = self.connection.cursor()
        cursor.execute(
            """
            SELECT datname
            FROM pg_database
            ORDER BY datname ASC
            """
        )
        return [row[0] for row in cursor.fetchall()]

    def schema(self, catalog: str, schema: str, table: str) -> StructType:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
                SELECT
                    information_schema.columns.*,
                    pg_attribute.attndims,
                    enums.enum_values
                FROM information_schema.columns

                -- Join to get the array dimensions
                JOIN pg_catalog.pg_namespace
                    ON pg_catalog.pg_namespace.nspname = information_schema.columns.table_schema
                JOIN pg_catalog.pg_class
                    ON pg_catalog.pg_class.relname = information_schema.columns.table_name
                    AND pg_catalog.pg_class.relnamespace = pg_catalog.pg_namespace.oid
                JOIN pg_catalog.pg_attribute
                    ON pg_catalog.pg_attribute.attrelid = pg_catalog.pg_class.oid
                    AND pg_catalog.pg_attribute.attname = information_schema.columns.column_name

                -- Join to get the enum values
                LEFT JOIN pg_catalog.pg_type
                    ON pg_catalog.pg_type.oid = pg_catalog.pg_attribute.atttypid
                    AND pg_catalog.pg_type.typtype = 'e' -- Ensuring it's an enum type
                LEFT JOIN (
                    SELECT
                        enumtypid,
                        array_agg(enumlabel) AS enum_values
                    FROM pg_catalog.pg_enum
                    GROUP BY enumtypid
                ) enums ON enums.enumtypid = pg_catalog.pg_type.oid
                WHERE table_name = {self.param_style}
                    AND table_schema = {self.param_style}
                    AND table_catalog = {self.param_style}
                ORDER BY ordinal_position ASC
            """,
            (table, schema, catalog),
        )
        names = [name[0].upper() for name in cursor.description]
        return self.converter.to_recap(
            # Make each row be a dict with the column names as keys
            [dict(zip(names, row)) for row in cursor.fetchall()]
        )



import pytest

from recap.converters.sqlite import SQLiteConverter
from recap.types import BytesType, FloatType, IntType, NullType, StringType, UnionType


# Test cases for SQLiteConverter
@pytest.mark.parametrize(
    "column_props,expected",
    [
        # INTEGER affinity tests
        (
            {
                "COLUMN_NAME": "id",
                "TYPE": "integer",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=64),
        ),
        (
            {
                "COLUMN_NAME": "big_id",
                "TYPE": "bigint",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=64),
        ),
        (
            {
                "COLUMN_NAME": "unsigned_big_id",
                "TYPE": "unsigned bigint",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            IntType(bits=64),
        ),
        # REAL affinity tests
        (
            {
                "COLUMN_NAME": "real_col",
                "TYPE": "real",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 22,
                "NUMERIC_SCALE": None,
            },
            FloatType(bits=32),
        ),
        (
            {
                "COLUMN_NAME": "float_col",
                "TYPE": "float",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 53,
                "NUMERIC_SCALE": None,
            },
            FloatType(bits=64),
        ),
        # TEXT affinity tests
        (
            {
                "COLUMN_NAME": "text_col",
                "TYPE": "text",
                "CHARACTER_OCTET_LENGTH": 255,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=255),
        ),
        (
            {
                "COLUMN_NAME": "varchar_col",
                "TYPE": "varchar(100)",
                "CHARACTER_OCTET_LENGTH": 100,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            StringType(bytes_=100),
        ),
        # BLOB affinity tests
        (
            {
                "COLUMN_NAME": "blob_col",
                "TYPE": "blob",
                "CHARACTER_OCTET_LENGTH": 500,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            BytesType(bytes_=500),
        ),
        # NUMERIC affinity tests
        (
            {
                "COLUMN_NAME": "numeric_col",
                "TYPE": "numeric",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 5,
            },
            UnionType(
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ]
            ),
        ),
        (
            {
                "COLUMN_NAME": "decimal_col",
                "TYPE": "decimal(10,5)",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 5,
            },
            UnionType(
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ]
            ),
        ),
        # Edge case tests
        (
            {
                "COLUMN_NAME": "custom_col",
                "TYPE": "customtype",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            UnionType(
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ]
            ),
        ),
        (
            {
                "COLUMN_NAME": "no_type_col",
                "TYPE": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            BytesType(bytes_=2147483647),
        ),
        # Date and time
        (
            {
                "COLUMN_NAME": "date_col",
                "TYPE": "DATE",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            UnionType(
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ]
            ),
        ),
        (
            {
                "COLUMN_NAME": "datetime_col",
                "TYPE": "DATETIME",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            UnionType(
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ]
            ),
        ),
        (
            {
                "COLUMN_NAME": "timestamp_col",
                "TYPE": "TIMESTAMP",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            UnionType(
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ]
            ),
        ),
        (
            {
                "COLUMN_NAME": "time_col",
                "TYPE": "TIME",
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
            },
            UnionType(
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ]
            ),
        ),
    ],
)
def test_sqlite_converter(column_props, expected):
    result = SQLiteConverter()._parse_type(column_props)
    assert result == expected



import sqlite3
from typing import Generator

import pytest

from recap.clients.sqlite import SQLiteClient
from recap.types import (
    BytesType,
    FloatType,
    IntType,
    NullType,
    StringType,
    StructType,
    UnionType,
)


@pytest.fixture
def sqlite_client() -> Generator[SQLiteClient, None, None]:
    with sqlite3.connect(":memory:") as connection:
        cursor = connection.cursor()
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS test_affinities (
                id INTEGER PRIMARY KEY,
                name TEXT NOT NULL,
                short_name CHAR(20) NOT NULL,
                variable_name VARCHAR(255) NOT NULL,
                descriptive_text CLOB NOT NULL,
                balance REAL,
                small_balance REAL(5, 2),
                small_dollars FLOAT(5),
                significant_figures DOUBLE,
                data BLOB,
                small_data BLOB(255),
                custom_type FOO,
                date_col DATE NOT NULL,
                datetime_col DATETIME NOT NULL,
                time_col TIME NOT NULL,
                timestamp_col TIMESTAMP NOT NULL
            );
            """
        )
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS test_strict (
                id INTEGER PRIMARY KEY,
                name TEXT NOT NULL,
                age INT NOT NULL,
                balance REAL NOT NULL,
                data BLOB NOT NULL,
                anything ANY,
                nullable_int INT
            ) STRICT;
            """
        )
        connection.commit()
        yield SQLiteClient(connection)  # type: ignore


def test_schema_affinities(sqlite_client: SQLiteClient):
    schema = sqlite_client.schema("test_affinities")

    expected_schema = StructType(
        fields=[
            IntType(bits=64, signed=True, name="id"),
            StringType(bytes_=2147483647, name="name", variable=True),
            StringType(bytes_=20, name="short_name", variable=True),
            StringType(bytes_=255, name="variable_name", variable=True),
            StringType(bytes_=2147483647, name="descriptive_text", variable=True),
            UnionType(
                types=[NullType(), FloatType(bits=64)],
                name="balance",
                default=None,
            ),
            UnionType(
                types=[NullType(), FloatType(bits=32)],
                name="small_balance",
                default=None,
            ),
            UnionType(
                types=[NullType(), FloatType(bits=32)],
                name="small_dollars",
                default=None,
            ),
            UnionType(
                types=[NullType(), FloatType(bits=64)],
                name="significant_figures",
                default=None,
            ),
            UnionType(
                types=[NullType(), BytesType(bytes_=2147483647)],
                name="data",
                default=None,
            ),
            UnionType(
                types=[NullType(), BytesType(bytes_=255)],
                name="small_data",
                default=None,
            ),
            UnionType(
                default=None,
                name="custom_type",
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ],
            ),
            UnionType(
                name="date_col",
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ],
            ),
            UnionType(
                name="datetime_col",
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ],
            ),
            UnionType(
                name="time_col",
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ],
            ),
            UnionType(
                name="timestamp_col",
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ],
            ),
        ]
    )

    assert len(schema.fields) == len(expected_schema.fields)
    for field, expected_field in zip(schema.fields, expected_schema.fields):
        assert field == expected_field


def test_schemas_strict(sqlite_client: SQLiteClient):
    schema = sqlite_client.schema("test_strict")

    expected_schema = StructType(
        fields=[
            IntType(bits=64, signed=True, name="id"),
            StringType(bytes_=2147483647, name="name", variable=True),
            IntType(bits=64, signed=True, name="age"),
            FloatType(bits=64, name="balance"),
            BytesType(bytes_=2147483647, name="data"),
            UnionType(
                types=[
                    NullType(),
                    IntType(bits=64),
                    FloatType(bits=64),
                    StringType(bytes_=2147483647),
                    BytesType(bytes_=2147483647),
                ],
                name="anything",
                default=None,
            ),
            UnionType(
                types=[NullType(), IntType(bits=64)],
                name="nullable_int",
                default=None,
            ),
        ]
    )

    assert len(schema.fields) == len(expected_schema.fields)
    for field, expected_field in zip(schema.fields, expected_schema.fields):
        assert field == expected_field


def test_ls(sqlite_client: SQLiteClient):
    assert set(sqlite_client.ls()) == {"test_affinities", "test_strict"}



from __future__ import annotations

from contextlib import contextmanager
from re import compile as re_compile
from typing import Any, Generator

from recap.clients.dbapi import Connection
from recap.converters.sqlite import SQLiteAffinity, SQLiteConverter
from recap.types import StructType

SQLITE3_CONNECT_ARGS = {
    "database",
    "timeout",
    "detect_types",
    "isolation_level",
    "check_same_thread",
    "factory",
    "cached_statements",
    "uri",
}


class SQLiteClient:
    def __init__(self, connection: Connection) -> None:
        self.connection = connection
        self.converter = SQLiteConverter()

    @staticmethod
    @contextmanager
    def create(url: str, **url_args) -> Generator[SQLiteClient, None, None]:
        import sqlite3

        # Strip sqlite:/// URL prefix
        url_args["database"] = url[len("sqlite:///") :]

        # Only include kwargs that are valid for PsycoPG2 parse_dsn()
        url_args = {k: v for k, v in url_args.items() if k in SQLITE3_CONNECT_ARGS}

        with sqlite3.connect(**url_args) as client:
            yield SQLiteClient(client)  # type: ignore

    @staticmethod
    def parse(method: str, **url_args) -> tuple[str, list[Any]]:
        from urllib.parse import urlunparse

        match method:
            case "ls":
                return (url_args["url"], [])
            case "schema":
                table = url_args["paths"].pop(-1)
                connection_url = urlunparse(
                    [
                        url_args.get("dialect") or url_args.get("scheme"),
                        url_args.get("netloc"),
                        # Include / prefix for paths
                        "/".join(url_args.get("paths", [])),
                        url_args.get("params"),
                        url_args.get("query"),
                        url_args.get("fragment"),
                    ]
                )

                # urlunsplit does not double slashes if netloc is empty. But most
                # URLs with empty netloc should have a double slash (e.g.
                # bigquery:// or sqlite:///some/file.db). Include an extra "/"
                # because the root path is not included with an empty netloc
                # and join().
                if not url_args.get("netloc"):
                    connection_url = connection_url.replace(":", ":///", 1)

                return (connection_url, [table])
            case _:
                raise ValueError("Invalid method")

    def ls(self) -> list[str]:
        cursor = self.connection.cursor()
        cursor.execute("SELECT name FROM sqlite_schema WHERE type='table'")
        return [row[0] for row in cursor.fetchall()]

    def schema(self, table: str) -> StructType:
        cursor = self.connection.cursor()

        # Validate that table exists since we want to prevent SQL injections in
        # the PRAGMA call
        if not self._table_exists(table):
            raise ValueError(f"Table '{table}' does not exist in the database.")

        cursor.execute(f"PRAGMA table_info({table});")
        names = [name[0].upper() for name in cursor.description]
        rows = []

        for row_cells in cursor.fetchall():
            row = dict(zip(names, row_cells))
            row = self.add_information_schema(row)
            rows.append(row)

        return self.converter.to_recap(rows)

    def add_information_schema(self, row: dict[str, Any]) -> dict[str, Any]:
        """
        SQLite does not have an INFORMATION_SCHEMA, so we need to add these
        columns.

        :param row: A row from the PRAGMA table_info() query.
        :return: The row with the INFORMATION_SCHEMA columns added.
        """

        is_not_null = row["NOTNULL"] or row["PK"]

        # Set defaults.
        information_schema_cols = {
            "COLUMN_NAME": row["NAME"],
            "IS_NULLABLE": "NO" if is_not_null else "YES",
            "COLUMN_DEFAULT": row["DFLT_VALUE"],
            "NUMERIC_PRECISION": None,
            "NUMERIC_SCALE": None,
            "CHARACTER_OCTET_LENGTH": None,
        }

        # Extract precision, scale, and octet length.
        # This regex matches the following patterns:
        # - <type>(<param1>(, <param2>)?)
        # param1 can be a precision or octet length, and param2 can be a scale.
        numeric_pattern = re_compile(r"(\w+)\((\d+)(?:,\s*(\d+))?\)")
        param_match = numeric_pattern.search(row["TYPE"])

        if param_match:
            # Extract matched values
            base_type, precision, scale = param_match.groups()
            base_type = base_type.upper()
            precision = int(precision)
            scale = int(scale) if scale else 0

            match SQLiteConverter.get_affinity(base_type):
                case SQLiteAffinity.INTEGER | SQLiteAffinity.REAL | SQLiteAffinity.NUMERIC:
                    information_schema_cols["NUMERIC_PRECISION"] = precision
                    information_schema_cols["NUMERIC_SCALE"] = scale
                case SQLiteAffinity.TEXT | SQLiteAffinity.BLOB:
                    information_schema_cols["CHARACTER_OCTET_LENGTH"] = precision

        return row | information_schema_cols

    def _table_exists(self, table: str) -> bool:
        cursor = self.connection.cursor()
        cursor.execute(
            "SELECT name FROM sqlite_master WHERE type='table' AND name=?", (table,)
        )
        return bool(cursor.fetchone())



from unittest.mock import MagicMock

import pytest
from confluent_kafka import schema_registry

from recap.clients.confluent_registry import ConfluentRegistryClient
from recap.types import IntType, StringType, StructType, UnionType


def test_struct_avro():
    mock_schema_registry_client = MagicMock(spec=schema_registry.SchemaRegistryClient)

    class MockSchema:
        schema_type = "AVRO"
        schema_str = """
        {
          "type": "record",
          "name": "User",
          "fields": [
            {"name": "name", "type": "string"},
            {"name": "age", "type": "int"}
          ]
        }
        """

    class MockRegisteredSchema:
        schema = MockSchema()

    mock_schema_registry_client.get_latest_version.return_value = MockRegisteredSchema()

    client = ConfluentRegistryClient(mock_schema_registry_client)
    result = client.schema("dummy_topic-value")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert result.fields[0].extra_attrs["name"] == "name"
    assert isinstance(result.fields[0], StringType)
    assert result.fields[1].extra_attrs["name"] == "age"
    assert isinstance(result.fields[1], IntType)

    # Check that the get_latest_version method was called with the correct subject.
    mock_schema_registry_client.get_latest_version.assert_called_with(
        "dummy_topic-value"
    )


def test_struct_protobuf():
    mock_schema_registry_client = MagicMock(spec=schema_registry.SchemaRegistryClient)

    class MockSchema:
        schema_type = "PROTOBUF"
        schema_str = """
        syntax = "proto3";

        message Person {
          string name = 1;
          int32 age = 2;
        }
        """

    class MockRegisteredSchema:
        schema = MockSchema()

    mock_schema_registry_client.get_latest_version.return_value = MockRegisteredSchema()

    client = ConfluentRegistryClient(mock_schema_registry_client)
    result = client.schema("dummy_topic-value")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert isinstance(result.fields[0], UnionType)
    assert isinstance(result.fields[0].types[1], StringType)
    assert isinstance(result.fields[1], UnionType)
    assert isinstance(result.fields[1].types[1], IntType)

    # Check that the get_latest_version method was called with the correct subject.
    mock_schema_registry_client.get_latest_version.assert_called_with(
        "dummy_topic-value"
    )


def test_struct_json_schema():
    mock_schema_registry_client = MagicMock(spec=schema_registry.SchemaRegistryClient)

    class MockSchema:
        schema_type = "JSON"
        schema_str = """
        {
          "$schema": "http://json-schema.org/draft-07/schema#",
          "title": "User",
          "type": "object",
          "properties": {
            "name": {
                "type": "string"
            },
            "age": {
                "type": "integer"
            }
          },
          "required": ["name", "age"]
        }
        """

    class MockRegisteredSchema:
        schema = MockSchema()

    mock_schema_registry_client.get_latest_version.return_value = MockRegisteredSchema()

    client = ConfluentRegistryClient(mock_schema_registry_client)
    result = client.schema("dummy_topic-value")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert result.fields[0].extra_attrs["name"] == "name"
    assert isinstance(result.fields[0], StringType)
    assert result.fields[1].extra_attrs["name"] == "age"
    assert isinstance(result.fields[1], IntType)

    # Check that the get_latest_version method was called with the correct subject.
    mock_schema_registry_client.get_latest_version.assert_called_with(
        "dummy_topic-value"
    )


@pytest.mark.parametrize(
    "method, url_args, expected_result",
    [
        # Test ls method
        ("ls", {"url": "http+csr://registry-url"}, ("http://registry-url", [])),
        (
            "ls",
            {
                "url": "http+csr://registry-url/path1/path2?ssl.ca.location=/foo/bar.crt#foo",
                "paths": ["path1", "path2"],
                "scheme": "http+csr",
                "dialect": "http",
                "netloc": "registry-url",
                "query": "ssl.ca.location=/foo/bar.crt",
                "fragment": "foo",
            },
            ("http://registry-url/path1/path2?ssl.ca.location=/foo/bar.crt#foo", []),
        ),
        # Test schema method
        (
            "schema",
            {
                "url": "http+csr://registry-url/subject",
                "paths": ["subject"],
                "scheme": "http+csr",
                "dialect": "http",
                "netloc": "registry-url",
            },
            ("http://registry-url", ["subject"]),
        ),
        (
            "schema",
            {
                "url": "http+csr://registry-url/path1/path2/subject",
                "paths": ["path1", "path2", "subject"],
                "scheme": "http+csr",
                "dialect": "http",
                "netloc": "registry-url",
                "query": "ssl.ca.location=/foo/bar.crt",
                "fragment": "foo",
            },
            (
                "http://registry-url/path1/path2?ssl.ca.location=/foo/bar.crt#foo",
                ["subject"],
            ),
        ),
        # Test invalid method
        (
            "invalid_method",
            {"url": "http+csr://registry-url"},
            pytest.raises(ValueError, match="Invalid method"),
        ),
    ],
)
def test_parse_method(method, url_args, expected_result):
    if isinstance(expected_result, tuple):
        result = ConfluentRegistryClient.parse(method, **url_args)
        assert result == expected_result
    else:
        with expected_result:
            ConfluentRegistryClient.parse(method, **url_args)



import psycopg2

from recap.clients import create_client
from recap.clients.postgresql import PostgresqlClient
from recap.converters.postgresql import MAX_FIELD_SIZE, PostgresqlConverter
from recap.types import (
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    NullType,
    ProxyType,
    RecapType,
    StringType,
    StructType,
    UnionType,
)


class TestPostgresqlClient:
    @classmethod
    def setup_class(cls):
        # Connect to the PostgreSQL database
        cls.connection = psycopg2.connect(
            host="localhost",
            port="5432",
            user="postgres",
            password="password",
            dbname="testdb",
        )

        # Create custom types
        cursor = cls.connection.cursor()
        cursor.execute(
            """
            CREATE TYPE test_enum_type_mood AS ENUM (
                'sad',
                'ok',
                'happy'
            );
            """
        )

        # Create tables
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS test_types (
                test_bigint BIGINT,
                test_integer INTEGER,
                test_smallint SMALLINT,
                test_float DOUBLE PRECISION,
                test_real REAL,
                test_boolean BOOLEAN,
                test_text TEXT,
                test_char CHAR(10),
                test_bytea BYTEA,
                test_bit BIT(10),
                test_timestamp TIMESTAMP,
                test_decimal DECIMAL(10,2),
                test_not_null INTEGER NOT NULL,
                test_not_null_default INTEGER NOT NULL DEFAULT 1,
                test_default INTEGER DEFAULT 2,
                test_int_array INTEGER[],
                test_varchar_array VARCHAR(255)[] DEFAULT '{"Hello", "World"}',
                test_bit_array BIT(8)[],
                test_not_null_array INTEGER[] NOT NULL,
                test_int_array_2d INTEGER[][],
                test_text_array_3d TEXT[][][],
                test_enum_mood test_enum_type_mood
            );
        """
        )
        cls.connection.commit()

    @classmethod
    def teardown_class(cls):
        # Delete tables
        cursor = cls.connection.cursor()
        cursor.execute("DROP TABLE IF EXISTS test_types;")
        cursor.execute("DROP TYPE IF EXISTS test_enum_type_mood;")
        cls.connection.commit()

        # Close the connection
        cls.connection.close()

    def test_struct_method_arrays_no_enforce_dimensions(self):
        client = PostgresqlClient(self.connection, PostgresqlConverter(False))
        test_types_struct = client.schema("testdb", "public", "test_types")

        expected_fields = [
            UnionType(
                default=None,
                name="test_bigint",
                types=[NullType(), IntType(bits=64, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_integer",
                types=[NullType(), IntType(bits=32, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_smallint",
                types=[NullType(), IntType(bits=16, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_float",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="test_real",
                types=[NullType(), FloatType(bits=32)],
            ),
            UnionType(
                default=None,
                name="test_boolean",
                types=[NullType(), BoolType()],
            ),
            UnionType(
                default=None,
                name="test_text",
                types=[NullType(), StringType(bytes_=MAX_FIELD_SIZE, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_char",
                # 40 = max of 4 bytes in a UTF-8 encoded unicode character * 10 chars
                types=[NullType(), StringType(bytes_=40, variable=False)],
            ),
            UnionType(
                default=None,
                name="test_bytea",
                types=[NullType(), BytesType(bytes_=MAX_FIELD_SIZE, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_bit",
                types=[NullType(), BytesType(bytes_=2, variable=False)],
            ),
            UnionType(
                default=None,
                name="test_timestamp",
                types=[
                    NullType(),
                    IntType(
                        bits=64, logical="build.recap.Timestamp", unit="microsecond"
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_decimal",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=32,
                        variable=False,
                        precision=10,
                        scale=2,
                    ),
                ],
            ),
            IntType(bits=32, signed=True, name="test_not_null"),
            IntType(bits=32, signed=True, name="test_not_null_default", default="1"),
            UnionType(
                default="2",
                name="test_default",
                types=[NullType(), IntType(bits=32, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_int_array",
                types=[
                    NullType(),
                    ListType(
                        alias="_root.test_int_array",
                        values=UnionType(
                            types=[
                                IntType(bits=32),
                                ProxyType(
                                    alias="_root.test_int_array",
                                    registry=client.converter.registry,  # type: ignore
                                ),
                            ]
                        ),
                    ),
                ],
            ),
            UnionType(
                default="'{Hello,World}'::character varying[]",
                name="test_varchar_array",
                types=[
                    NullType(),
                    ListType(
                        alias="_root.test_varchar_array",
                        values=UnionType(
                            types=[
                                StringType(bytes_=MAX_FIELD_SIZE),
                                ProxyType(
                                    alias="_root.test_varchar_array",
                                    registry=client.converter.registry,  # type: ignore
                                ),
                            ]
                        ),
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_bit_array",
                types=[
                    NullType(),
                    ListType(
                        alias="_root.test_bit_array",
                        values=UnionType(
                            types=[
                                BytesType(bytes_=MAX_FIELD_SIZE, variable=False),
                                ProxyType(
                                    alias="_root.test_bit_array",
                                    registry=client.converter.registry,  # type: ignore
                                ),
                            ]
                        ),
                    ),
                ],
            ),
            ListType(
                name="test_not_null_array",
                alias="_root.test_not_null_array",
                values=UnionType(
                    types=[
                        IntType(bits=32),
                        ProxyType(
                            alias="_root.test_not_null_array",
                            registry=client.converter.registry,  # type: ignore
                        ),
                    ]
                ),
            ),
            UnionType(
                default=None,
                name="test_int_array_2d",
                types=[
                    NullType(),
                    ListType(
                        alias="_root.test_int_array_2d",
                        values=UnionType(
                            types=[
                                IntType(bits=32),
                                ProxyType(
                                    alias="_root.test_int_array_2d",
                                    registry=client.converter.registry,  # type: ignore
                                ),
                            ]
                        ),
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_text_array_3d",
                types=[
                    NullType(),
                    ListType(
                        alias="_root.test_text_array_3d",
                        values=UnionType(
                            types=[
                                StringType(bytes_=MAX_FIELD_SIZE, variable=True),
                                ProxyType(
                                    alias="_root.test_text_array_3d",
                                    registry=client.converter.registry,  # type: ignore
                                ),
                            ]
                        ),
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_enum_mood",
                types=[
                    NullType(),
                    EnumType(symbols=["sad", "ok", "happy"]),
                ],
            ),
        ]
        validate_results(test_types_struct, expected_fields)

    def test_struct_method_arrays_enforce_dimensions(self):
        client = PostgresqlClient(self.connection, PostgresqlConverter(True))  # type: ignore
        test_types_struct = client.schema("testdb", "public", "test_types")

        expected_fields = [
            UnionType(
                default=None,
                name="test_bigint",
                types=[NullType(), IntType(bits=64, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_integer",
                types=[NullType(), IntType(bits=32, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_smallint",
                types=[NullType(), IntType(bits=16, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_float",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="test_real",
                types=[NullType(), FloatType(bits=32)],
            ),
            UnionType(
                default=None,
                name="test_boolean",
                types=[NullType(), BoolType()],
            ),
            UnionType(
                default=None,
                name="test_text",
                types=[NullType(), StringType(bytes_=MAX_FIELD_SIZE, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_char",
                # 40 = max of 4 bytes in a UTF-8 encoded unicode character * 10 chars
                types=[NullType(), StringType(bytes_=40, variable=False)],
            ),
            UnionType(
                default=None,
                name="test_bytea",
                types=[NullType(), BytesType(bytes_=MAX_FIELD_SIZE, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_bit",
                types=[NullType(), BytesType(bytes_=2, variable=False)],
            ),
            UnionType(
                default=None,
                name="test_timestamp",
                types=[
                    NullType(),
                    IntType(
                        bits=64, logical="build.recap.Timestamp", unit="microsecond"
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_decimal",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=32,
                        variable=False,
                        precision=10,
                        scale=2,
                    ),
                ],
            ),
            IntType(bits=32, signed=True, name="test_not_null"),
            IntType(bits=32, signed=True, name="test_not_null_default", default="1"),
            UnionType(
                default="2",
                name="test_default",
                types=[NullType(), IntType(bits=32, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_int_array",
                types=[
                    NullType(),
                    ListType(
                        values=UnionType(
                            types=[
                                NullType(),
                                IntType(bits=32),
                            ]
                        ),
                    ),
                ],
            ),
            UnionType(
                default="'{Hello,World}'::character varying[]",
                name="test_varchar_array",
                types=[
                    NullType(),
                    ListType(
                        values=UnionType(
                            types=[
                                NullType(),
                                StringType(bytes_=MAX_FIELD_SIZE),
                            ]
                        ),
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_bit_array",
                types=[
                    NullType(),
                    ListType(
                        values=UnionType(
                            types=[
                                NullType(),
                                BytesType(bytes_=MAX_FIELD_SIZE, variable=False),
                            ]
                        ),
                    ),
                ],
            ),
            ListType(
                name="test_not_null_array",
                values=UnionType(
                    types=[
                        NullType(),
                        IntType(bits=32),
                    ]
                ),
            ),
            UnionType(
                default=None,
                name="test_int_array_2d",
                types=[
                    NullType(),
                    ListType(
                        values=ListType(
                            values=UnionType(
                                types=[
                                    NullType(),
                                    IntType(bits=32),
                                ]
                            )
                        ),
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_text_array_3d",
                types=[
                    NullType(),
                    ListType(
                        values=ListType(
                            values=ListType(
                                values=UnionType(
                                    types=[
                                        NullType(),
                                        StringType(
                                            bytes_=MAX_FIELD_SIZE, variable=True
                                        ),
                                    ]
                                )
                            )
                        ),
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_enum_mood",
                types=[
                    NullType(),
                    EnumType(symbols=["sad", "ok", "happy"]),
                ],
            ),
        ]
        validate_results(test_types_struct, expected_fields)

    def test_create_client(self):
        postgresql_url = "postgresql://postgres:password@localhost:5432/testdb"

        with create_client(postgresql_url) as client:
            assert client.ls() == ["postgres", "template0", "template1", "testdb"]
            assert client.ls("testdb") == [
                "pg_toast",
                "pg_catalog",
                "public",
                "information_schema",
            ]
            assert client.ls("testdb", "public") == ["test_types"]


def validate_results(
    test_types_struct: StructType, expected_fields: list[RecapType]
) -> None:
    # Going field by field to make debugging easier when test fails
    for field, expected_field in zip(test_types_struct.fields, expected_fields):
        assert field == expected_field

    assert test_types_struct == StructType(fields=expected_fields)
